from datetime import datetime

from sqlalchemy import Column, DateTime
from sqlalchemy.event import listen

from cytra.db.created_mixin import CreatedMixin


class ModifiedMixin(CreatedMixin):
    modified_at = Column(DateTime, nullable=True)

    @property
    def last_modification_time(self):
        return self.modified_at or self.created_at

    @staticmethod
    def before_update(mapper, connection, target):
        target.modified_at = datetime.utcnow()

    @classmethod
    def __declare_last__(cls):
        listen(cls, "before_update", cls.before_update)



from redis import Redis


class RedisAppMixin:
    redis = None

    def setup(self):
        if "redis" in self.config:
            self.redis = Redis(**dict(self.config.redis))
        else:
            self.redis = None
        super().setup()

    def shutdown(self):
        super().shutdown()
        if "redis" in self.config:
            self.redis.close()



from logging import getLogger, shutdown
from logging.config import dictConfig


class LogAppMixin:
    _logger_name = "main"

    def setup(self):
        if "logging" in self.config:
            dictConfig(dict(self.config.logging))
            self.log = getLogger(self._logger_name)
        super().setup()

    def shutdown(self):
        super().shutdown()
        if "logging" in self.config:
            shutdown()
            del self.log



from gongish import Application as GongishApp

from cytra.auth import AuthAppMixin
from cytra.cors import CORS
from cytra.db import DatabaseAppMixin
from cytra.log import LogAppMixin
from cytra.redis import RedisAppMixin


class CytraAppBase(GongishApp):
    # Reserved CORS config
    cors: CORS = None

    @staticmethod
    def format_json(request, response, indent=None):
        if hasattr(response.body, "to_dict"):
            response.body = response.body.to_dict()
        elif hasattr(response.body, "expose"):
            response.body = response.body.expose()
        GongishApp.format_json(request, response, indent)

    default_formatter = format_json

    def __init__(self):
        super().__init__()
        self.cors = CORS()
        self.cors.allow_headers.add("content-type")
        self.cors.expose_headers.add("content-type")

    def setup(self):
        pass

    def shutdown(self):
        pass


class Application(
    AuthAppMixin,
    RedisAppMixin,
    DatabaseAppMixin,
    LogAppMixin,
    CytraAppBase,
):
    pass



from datetime import datetime

from sqlalchemy import Column, DateTime


class CreatedMixin:
    created_at = Column(
        DateTime,
        default=lambda: datetime.utcnow(),
        nullable=False,
    )



# flake8: noqa
# isort: skip_file
from cytra.auth.authenticator import Authenticator
from cytra.auth.principal import DefaultJWTPrincipal, JWTPrincipal
from cytra.auth.stateful_authenticator import StatefulAuthenticator
from cytra.auth.app_mixin import AuthAppMixin



from gongish import HTTPBadRequest, HTTPStatus


class CytraException(Exception):
    pass


class InvalidParamError(HTTPBadRequest, CytraException):
    pass


class PaginationLimitError(HTTPBadRequest, CytraException):
    text = "Pagination limit error"


class PostgresSQLError(HTTPStatus, CytraException):
    def __init__(self, exc, *args):
        error_code = exc.orig.pgcode
        self.code = self.__class__.pg_status_code_map.get(error_code, 500)
        self.text = self.__class__.pg_status_text_map.get(error_code)
        super().__init__(*args)

    @staticmethod
    def validate_postgres_error(exc):
        return (
            hasattr(exc, "orig")
            and exc.orig is not None
            and hasattr(exc.orig, "pgcode")
        )

    pg_status_code_map = {
        "23502": 400,
        "23505": 409,
        "22P02": 400,
        "23503": 400,
    }
    pg_status_text_map = {
        "00000": "successful_completion",
        "01000": "warning",
        "0100C": "dynamic_result_sets_returned",
        "01008": "implicit_zero_bit_padding",
        "01003": "null_value_eliminated_in_set_function",
        "01007": "privilege_not_granted",
        "01006": "privilege_not_revoked",
        "01004": "string_data_right_truncation",
        "01P01": "deprecated_feature",
        "02000": "no_data",
        "02001": "no_additional_dynamic_result_sets_returned",
        "03000": "sql_statement_not_yet_complete",
        "08000": "connection_exception",
        "08003": "connection_does_not_exist",
        "08006": "connection_failure",
        "08001": "sqlclient_unable_to_establish_sqlconnection",
        "08004": "sqlserver_rejected_establishment_of_sqlconnection",
        "08007": "transaction_resolution_unknown",
        "08P01": "protocol_violation",
        "09000": "triggered_action_exception",
        "0A000": "feature_not_supported",
        "0B000": "invalid_transaction_initiation",
        "0F000": "locator_exception",
        "0F001": "invalid_locator_specification",
        "0L000": "invalid_grantor",
        "0LP01": "invalid_grant_operation",
        "0P000": "invalid_role_specification",
        "20000": "case_not_found",
        "21000": "cardinality_violation",
        "22000": "data_exception",
        "2202E": "array_subscript_error",
        "22021": "character_not_in_repertoire",
        "22008": "datetime_field_overflow",
        "22012": "division_by_zero",
        "22005": "error_in_assignment",
        "2200B": "escape_character_conflict",
        "22022": "indicator_overflow",
        "22015": "interval_field_overflow",
        "2201E": "invalid_argument_for_logarithm",
        "22014": "invalid_argument_for_ntile_function",
        "22016": "invalid_argument_for_nth_value_function",
        "2201F": "invalid_argument_for_power_function",
        "2201G": "invalid_argument_for_width_bucket_function",
        "22018": "invalid_character_value_for_cast",
        "22007": "invalid_datetime_format",
        "22019": "invalid_escape_character",
        "2200D": "invalid_escape_octet",
        "22025": "invalid_escape_sequence",
        "22P06": "nonstandard_use_of_escape_character",
        "22010": "invalid_indicator_parameter_value",
        "22023": "invalid_parameter_value",
        "2201B": "invalid_regular_expression",
        "2201W": "invalid_row_count_in_limit_clause",
        "2201X": "invalid_row_count_in_result_offset_clause",
        "22009": "invalid_time_zone_displacement_value",
        "2200C": "invalid_use_of_escape_character",
        "2200G": "most_specific_type_mismatch",
        "22004": "null_value_not_allowed",
        "22002": "null_value_no_indicator_parameter",
        "22003": "numeric_value_out_of_range",
        "22026": "string_data_length_mismatch",
        "22001": "string_data_right_truncation",
        "22011": "substring_error",
        "22027": "trim_error",
        "22024": "unterminated_c_string",
        "2200F": "zero_length_character_string",
        "22P01": "floating_point_exception",
        "22P02": "invalid_text_representation",
        "22P03": "invalid_binary_representation",
        "22P04": "bad_copy_file_format",
        "22P05": "untranslatable_character",
        "2200L": "not_an_xml_document",
        "2200M": "invalid_xml_document",
        "2200N": "invalid_xml_content",
        "2200S": "invalid_xml_comment",
        "2200T": "invalid_xml_processing_instruction",
        "23000": "integrity_constraint_violation",
        "23001": "restrict_violation",
        "23502": "not_null_violation",
        "23503": "foreign_key_violation",
        "23505": "unique_violation",
        "23514": "check_violation",
        "23P01": "exclusion_violation",
        "24000": "invalid_cursor_state",
        "25000": "invalid_transaction_state",
        "25001": "active_sql_transaction",
        "25002": "branch_transaction_already_active",
        "25008": "held_cursor_requires_same_isolation_level",
        "25003": "inappropriate_access_mode_for_branch_transaction",
        "25004": "inappropriate_isolation_level_for_branch_transaction",
        "25005": "no_active_sql_transaction_for_branch_transaction",
        "25006": "read_only_sql_transaction",
        "25007": "schema_and_data_statement_mixing_not_supported",
        "25P01": "no_active_sql_transaction",
        "25P02": "in_failed_sql_transaction",
        "26000": "invalid_sql_statement_name",
        "27000": "triggered_data_change_violation",
        "28000": "invalid_authorization_specification",
        "28P01": "invalid_password",
        "2B000": "dependent_privilege_descriptors_still_exist",
        "2BP01": "dependent_objects_still_exist",
        "2D000": "invalid_transaction_termination",
        "2F000": "sql_routine_exception",
        "2F005": "function_executed_no_return_statement",
        "2F002": "modifying_sql_data_not_permitted",
        "2F003": "prohibited_sql_statement_attempted",
        "2F004": "reading_sql_data_not_permitted",
        "34000": "invalid_cursor_name",
        "38000": "external_routine_exception",
        "38001": "containing_sql_not_permitted",
        "38002": "modifying_sql_data_not_permitted",
        "38003": "prohibited_sql_statement_attempted",
        "38004": "reading_sql_data_not_permitted",
        "39000": "external_routine_invocation_exception",
        "39001": "invalid_sqlstate_returned",
        "39004": "null_value_not_allowed",
        "39P01": "trigger_protocol_violated",
        "39P02": "srf_protocol_violated",
        "3B000": "savepoint_exception",
        "3B001": "invalid_savepoint_specification",
        "3D000": "invalid_catalog_name",
        "3F000": "invalid_schema_name",
        "40000": "transaction_rollback",
        "40002": "transaction_integrity_constraint_violation",
        "40001": "serialization_failure",
        "40003": "statement_completion_unknown",
        "40P01": "deadlock_detected",
        "42000": "syntax_error_or_access_rule_violation",
        "42601": "syntax_error",
        "42501": "insufficient_privilege",
        "42846": "cannot_coerce",
        "42803": "grouping_error",
        "42P20": "windowing_error",
        "42P19": "invalid_recursion",
        "42830": "invalid_foreign_key",
        "42602": "invalid_name",
        "42622": "name_too_long",
        "42939": "reserved_name",
        "42804": "datatype_mismatch",
        "42P18": "indeterminate_datatype",
        "42P21": "collation_mismatch",
        "42P22": "indeterminate_collation",
        "42809": "wrong_object_type",
        "42703": "undefined_column",
        "42883": "undefined_function",
        "42P01": "undefined_table",
        "42P02": "undefined_parameter",
        "42704": "undefined_object",
        "42701": "duplicate_column",
        "42P03": "duplicate_cursor",
        "42P04": "duplicate_database",
        "42723": "duplicate_function",
        "42P05": "duplicate_prepared_statement",
        "42P06": "duplicate_schema",
        "42P07": "duplicate_table",
        "42712": "duplicate_alias",
        "42710": "duplicate_object",
        "42702": "ambiguous_column",
        "42725": "ambiguous_function",
        "42P08": "ambiguous_parameter",
        "42P09": "ambiguous_alias",
        "42P10": "invalid_column_reference",
        "42611": "invalid_column_definition",
        "42P11": "invalid_cursor_definition",
        "42P12": "invalid_database_definition",
        "42P13": "invalid_function_definition",
        "42P14": "invalid_prepared_statement_definition",
        "42P15": "invalid_schema_definition",
        "42P16": "invalid_table_definition",
        "42P17": "invalid_object_definition",
        "44000": "with_check_option_violation",
        "53000": "insufficient_resources",
        "53100": "disk_full",
        "53200": "out_of_memory",
        "53300": "too_many_connections",
        "54000": "program_limit_exceeded",
        "54001": "statement_too_complex",
        "54011": "too_many_columns",
        "54023": "too_many_arguments",
        "55000": "object_not_in_prerequisite_state",
        "55006": "object_in_use",
        "55P02": "cant_change_runtime_param",
        "55P03": "lock_not_available",
        "57000": "operator_intervention",
        "57014": "query_canceled",
        "57P01": "admin_shutdown",
        "57P02": "crash_shutdown",
        "57P03": "cannot_connect_now",
        "57P04": "database_dropped",
        "58030": "io_error",
        "58P01": "undefined_file",
        "58P02": "duplicate_file",
        "F0000": "config_file_error",
        "F0001": "lock_file_exists",
        "HV000": "fdw_error",
        "HV005": "fdw_column_name_not_found",
        "HV002": "fdw_dynamic_parameter_value_needed",
        "HV010": "fdw_function_sequence_error",
        "HV021": "fdw_inconsistent_descriptor_information",
        "HV024": "fdw_invalid_attribute_value",
        "HV007": "fdw_invalid_column_name",
        "HV008": "fdw_invalid_column_number",
        "HV004": "fdw_invalid_data_type",
        "HV006": "fdw_invalid_data_type_descriptors",
        "HV091": "fdw_invalid_descriptor_field_identifier",
        "HV00B": "fdw_invalid_handle",
        "HV00C": "fdw_invalid_option_index",
        "HV00D": "fdw_invalid_option_name",
        "HV090": "fdw_invalid_string_length_or_buffer_length",
        "HV00A": "fdw_invalid_string_format",
        "HV009": "fdw_invalid_use_of_null_pointer",
        "HV014": "fdw_too_many_handles",
        "HV001": "fdw_out_of_memory",
        "HV00P": "fdw_no_schemas",
        "HV00J": "fdw_option_name_not_found",
        "HV00K": "fdw_reply_handle",
        "HV00Q": "fdw_schema_not_found",
        "HV00R": "fdw_table_not_found",
        "HV00L": "fdw_unable_to_create_execution",
        "HV00M": "fdw_unable_to_create_reply",
        "HV00N": "fdw_unable_to_establish_connection",
        "P0000": "plpgsql_error",
        "P0001": "raise_exception",
        "P0002": "no_data_found",
        "P0003": "too_many_rows",
        "XX000": "internal_error",
        "XX001": "data_corrupted",
        "XX002": "index_corrupted",
    }



import re


def to_camel_case(text):
    """Transform snake_case to CamelCase"""
    return re.sub(r"(_\w)", lambda x: x.group(1)[1:].upper(), text)


def dot_object(obj):
    # Limited javascript-like dot-notation access for testing module
    # Python special keywords (for,while,True, etc) not works

    if isinstance(obj, dict):
        return DotDict(obj)

    if isinstance(obj, list):
        return DotList(obj)

    return obj


class DotList(list):
    def __getitem__(self, key):
        return dot_object(super().__getitem__(key))


class DotDict(dict):
    def __getitem__(self, key):
        return dot_object(super().__getitem__(key))

    def get(self, key, default=None):
        return dot_object(super().get(key, default))

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)



import re

ISO_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
ISO_DATETIME_PATTERN = re.compile(
    r"^(?P<datetime>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2})"
    r"(?:\.(\d*))?(Z|\+\d{2}:\d{2})?$"
)
ISO_DATE_FORMAT = "%Y-%m-%d"
ISO_TIME_FORMAT = "%H:%M:%S"



import functools

from gongish import HTTPForbidden, HTTPUnauthorized

from cytra.auth import Authenticator


class AuthAppMixin:
    authenticator = Authenticator
    auth = None

    def on_begin_request(self):
        self.identity = None

        if self.auth:
            self.auth.authenticate_request()

        super().on_begin_request()

    def on_end_response(self):
        super().on_end_response()
        self.identity = None

    @property
    def identity(self):
        return self.request and self.request.identity

    @identity.setter
    def identity(self, value):
        self.request.identity = value

    def setup(self):
        super().setup()
        if "auth" in self.config:
            self.auth = self.__class__.authenticator(app=self)
            self.cors.allow_headers.add("authorization")
            self.cors.expose_headers.add("x-new-jwt-token")
        else:
            self.auth = None

    def shutdown(self):
        if self.auth is not None:
            self.auth = None
        super().shutdown()

    def authorize(self, *roles):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):

                if not self.identity:
                    raise HTTPUnauthorized("No identity")

                if len(roles) > 0 and not self.identity.is_in_roles(*roles):
                    raise HTTPForbidden

                return func(*args, **kwargs)

            return wrapper

        if roles and callable(roles[0]):
            f = roles[0]
            roles = []
            return decorator(f)
        else:
            return decorator



from datetime import datetime

from sqlalchemy import Column, DateTime
from sqlalchemy.ext.hybrid import hybrid_property


class ActivationMixin:
    activated_at = Column(DateTime, nullable=True)

    @hybrid_property
    def is_active(self):
        return self.activated_at is not None

    @is_active.setter
    def is_active(self, value):
        self.activated_at = datetime.utcnow() if value else None

    @is_active.expression
    def is_active(self):
        return self.activated_at.isnot(None)

    @classmethod
    def filter_activated(cls, query=None):
        return (query or cls.query).filter(cls.is_active)



import functools
from urllib.parse import urlencode

import webtest

from cytra import Application
from cytra import json as jsonlib
from cytra.db import DatabaseManager, metadata
from cytra.helpers import dot_object


class WebtestResponse(webtest.TestResponse):
    @property
    def json(self):
        return dot_object(super().json)


class WebtestRequest(webtest.TestRequest):
    ResponseClass = WebtestResponse


class WebtestApp(webtest.TestApp):
    RequestClass = WebtestRequest


class TestingApp:
    __test__ = False  # Disable pytest collector

    def __init__(self, app: Application, config_filename: str):
        self.app = app
        app.configure(config_filename)
        if app.config.get("sqlalchemy") and app.config.get("sqlalchemy_test"):
            app.config.sqlalchemy = app.config.sqlalchemy_test
            self._db_enabled = True
        else:
            self._db_enabled = False

    def __enter__(self):
        if self._db_enabled:
            with DatabaseManager(
                engine_kwargs=self.app.config.sqlalchemy.engine,
                admin_engine_kwargs=self.app.config.sqlalchemy_admin.engine,
            ) as dbm:
                if dbm.database_exists():
                    dbm.drop_database()
                dbm.create_database()

        self.app.setup()

        if self._db_enabled:
            metadata.drop_all(bind=self.app.db.get_bind())
            metadata.create_all(bind=self.app.db.get_bind())

        self.testapp = WebtestApp(self.app)
        self.testapp.extra_environ["HTTP_ACCEPT"] = "application/json"
        self.testapp.extra_environ["REMOTE_ADDR"] = "127.0.0.1"
        self.testapp.extra_environ["HTTP_APP_CLIENT"] = "CytraApp-1.1.1-beta"
        self.testapp.extra_environ["HTTP_USER_AGENT"] = (
            "Mozilla/5.0 (iPhone; CPU iPhone OS 5_1 like Mac OS X) "
            "AppleWebKit/534.46 (KHTML, like Gecko) Version/5.1 "
            "Mobile/9B179 Safari/7534.48.3; fa-IR; some; extra; info)"
        )
        return self

    def __exit__(self, *args):
        self.app.shutdown()

    def get_access_token(self, as_):  # pragma: nocover
        raise NotImplementedError

    def authorize(self, as_: str = None):
        auth_key = "HTTP_AUTHORIZATION"
        if auth_key in self.testapp.extra_environ:
            del self.testapp.extra_environ[auth_key]

        if as_ is None or as_ == "visitor":
            return

        self.testapp.extra_environ[auth_key] = self.get_access_token(as_)

    def __call__(self, as_, path, method="get", json=None, qs=None, **kwargs):
        self.authorize(as_)

        if qs:
            path = "%s?%s" % (path, urlencode(qs))

        if json:
            kwargs.update(
                params=jsonlib.dumps(json),
                upload_files=None,
                content_type="application/json",
            )

        return self.testapp._gen_request(
            method.upper(), path, expect_errors=False, **kwargs
        )

    def __getattr__(self, key):
        if key not in self.app.verbs:
            return super().__getattribute__(key)

        return functools.partial(self.__call__, method=key)



from sqlalchemy import inspect

from cytra.db.access_control_mixin import InspectionMixin
from cytra.db.transform_mixin import TransformMixin


class AttrStateMixin(TransformMixin, InspectionMixin):
    def get_attr_changes(self) -> dict:
        state = inspect(self)
        changes = {}

        for attr in state.attrs:
            hist = state.get_history(attr.key, True)

            if not hist.has_changes():
                continue

            old_value = hist.deleted[0] if hist.deleted else None
            new_value = hist.added[0] if hist.added else None
            changes[attr.key] = (old_value, new_value)

        return changes

    def export_attr_changes(self) -> dict:
        res = {}
        inspections = self.get_inspections()["all_by_key"]
        for k, v in self.get_attr_changes().items():
            ic = inspections[k]
            res[self.export_column_name(k, ic.info)] = (
                self.export_value(ic.column, v[0]),
                self.export_value(ic.column, v[1]),
            )
        return res



# flake8: noqa
# isort: skip_file
from cytra.db.field import Field, relationship, synonym, composite

from cytra.db.inspection_mixin import InspectionMixin, InspectedColumn
from cytra.db.access_control_mixin import AccessControlMixin
from cytra.db.transform_mixin import TransformMixin
from cytra.db.serialize_mixin import SerializeMixin
from cytra.db.created_mixin import CreatedMixin
from cytra.db.modified_mixin import ModifiedMixin
from cytra.db.activation_mixin import ActivationMixin
from cytra.db.attrstate_mixin import AttrStateMixin
from cytra.db.filtering_mixin import FilteringMixin
from cytra.db.ordering_mixin import OrderingMixin
from cytra.db.pagination_mixin import PaginationMixin
from cytra.db.manager import DatabaseManager
from cytra.db.base import metadata, CytraDBQuery, DeclarativeBase, dbsession
from cytra.db.app_mixin import DatabaseAppMixin



# flake8: noqa
from cytra.application import Application
from cytra.cors import CORSAppMixin
from cytra.exceptions import CytraException, InvalidParamError
from cytra.i18n import I18nAppMixin
from cytra.testing import TestingApp

__version__ = "1.1.0"



from typing import Any, Iterable

from cytra.db import InspectedColumn, InspectionMixin


class AccessControlMixin(InspectionMixin):
    """
    Controls access to properties

    Note: All mixins must respect to this mixin while intracting with
    outside of the model.
    """

    @classmethod
    def get_readables(cls, model: Any = None) -> Iterable[InspectedColumn]:
        inspections = cls.get_inspections()
        excludes = set(
            map(cls.get_column_key, cls.get_excludes("read", model=model))
        )
        for ic in inspections["all"]:
            if (
                ic.type == "proxies"
                or ic.info.get("protected")
                or ic.key in excludes
            ):
                continue

            yield ic

    @classmethod
    def get_writables(cls, model: Any = None) -> Iterable[InspectedColumn]:
        inspections = cls.get_inspections()
        excludes = set(
            map(cls.get_column_key, cls.get_excludes("write", model=model))
        )
        for ic in inspections["all"]:
            if (
                ic.type == "relationships"
                or ic.info.get("readonly")
                or ic.key in excludes
            ):
                continue

            yield ic

    @classmethod
    def get_excludes(cls, mode: str, model: Any = None):
        if mode == "read":
            yield from cls.get_read_excludes(model)

        elif mode == "write":
            yield from cls.get_write_excludes(model)

    @classmethod
    def get_read_excludes(cls, model: Any = None):  # pragma: nocover
        return
        yield

    @classmethod
    def get_write_excludes(cls, model: Any = None):  # pragma: nocover
        return
        yield



import re
from datetime import datetime

import user_agents

from cytra import json
from cytra.auth import Authenticator


class StatefulAuthenticator(Authenticator):
    """
    Stateful Authenticator

    Redis data-model:
        sessions: HashMap { session_id: member_id }
        member_id: Set { session_id }
        agents: HashMap {
            session_id: {
                remoteAddress: 127.0.0.1,
                machine: pc,
                os: android 4.2.2,
                agent: chrome 55,
                client: CytraClient-js 3.4.5-alpha14,
                lastActivity: 2015-10-26T07:46:36.615661
            }
        }
    """

    remote_address_key = "REMOTE_ADDR"
    agent_key = "HTTP_USER_AGENT"
    client_name = "CytraClient"
    members_key = "auth:member:%s"
    sessions_key = "auth:sessions"
    session_info_key = "auth:sessions:%s:info"

    def __init__(self, app) -> None:
        super().__init__(app)
        if not app.redis:
            raise RuntimeError("cytra.redis not configured")

        self.redis = self.app.redis

    @classmethod
    def get_member_sessions_key(cls, member_id):
        return cls.members_key % member_id

    @classmethod
    def get_session_info_key(cls, session_id):
        return cls.session_info_key % session_id

    def extract_session_info(self):
        remote_address = machine = os = agent = client = app = None
        environ = self.app.request.environ
        if (
            self.remote_address_key in environ
            and environ[self.remote_address_key]
        ):
            remote_address = environ[self.remote_address_key]

        if self.agent_key in environ:
            agent_string = environ[self.agent_key]
            user_agent = user_agents.parse(agent_string)

            machine = user_agent.is_pc and "PC" or user_agent.device.family
            os = " ".join(
                [user_agent.os.family, user_agent.os.version_string]
            ).strip()
            agent = " ".join(
                [user_agent.browser.family, user_agent.browser.version_string]
            ).strip()

            matched_client = re.match(
                ".*%s-(?P<type>.+)/(?P<version>.+) \\((?P<features>.+)\\).*"
                % self.client_name,
                agent_string,
            )
            if matched_client:
                matched_client = matched_client.groupdict()
                client_type = matched_client["type"]
                client_version = matched_client["version"]
                # exp: "cytra-js 1.2.3-rc10"
                client = f"{self.client_name}-{client_type} {client_version}"

                features = matched_client["features"].split(";")
                if len(features) >= 3:
                    # exp: "MobileToken (shark) 1.2.3"
                    app = "%s (%s) %s" % (
                        features[0].strip(),
                        features[1].strip(),
                        features[2].strip(),
                    )

        return {
            "remoteAddress": remote_address or "NA",
            "machine": machine or "Other",
            "os": os or "Other",
            "agent": agent or "Other",
            "client": client or "Unknown",
            "app": app or "Unknown",
            "lastActivity": datetime.utcnow().isoformat(),
        }

    def update_session_info(self, session_id):
        self.redis.set(
            self.get_session_info_key(session_id),
            json.dumps(self.extract_session_info()),
        )

    def register_session(self, member_id, session_id):
        self.redis.hset(self.sessions_key, session_id, member_id)
        self.redis.sadd(self.get_member_sessions_key(member_id), session_id)
        self.redis.set(
            self.get_session_info_key(session_id),
            json.dumps(self.extract_session_info()),
        )

    def unregister_session(self, session_id=None):
        session_id = session_id or self.app.identity.session_id
        member_id = self.redis.hget(self.sessions_key, session_id).decode()
        self.redis.srem(self.get_member_sessions_key(member_id), session_id)
        self.redis.hdel(self.sessions_key, session_id)
        self.redis.delete(self.get_session_info_key(session_id))

    def validate_session(self, session_id):
        return self.redis.hexists(self.sessions_key, session_id)

    def invalidate_member(self, member_id=None):
        """Remove member all sessions"""
        while True:
            session_id = self.redis.spop(
                self.get_member_sessions_key(member_id)
            )
            if not session_id:
                break
            self.redis.hdel(self.sessions_key, session_id)
            self.redis.delete(self.get_session_info_key(session_id))
        self.redis.delete(self.get_member_sessions_key(member_id))

    def authenticate_request(self):
        super().authenticate_request()
        if not self.app.identity:
            return

        if not self.validate_session(self.app.identity.session_id):
            self.app.identity = None
            return

        self.update_session_info(self.app.identity.session_id)

    def get_member_sessions(self, member_id: int):
        return list(
            map(
                lambda x: x.decode(),
                self.redis.smembers(self.get_member_sessions_key(member_id)),
            )
        )

    def get_session_info(self, session_id):
        info = self.redis.get(self.get_session_info_key(session_id))
        if info:
            return json.loads(info)
        return None



from typing import Union

import jwt
from gongish import HTTPBadRequest, HTTPUnauthorized

from cytra.auth.principal import DefaultJWTPrincipal, JWTPrincipal


class Authenticator:
    header_token_key = "HTTP_AUTHORIZATION"
    jwt_principal = DefaultJWTPrincipal

    def __init__(self, app) -> None:
        self.app = app

    def create_token(self, payload: dict):
        return JWTPrincipal(app=self.app, payload=payload).dump()

    def get_principal(self) -> Union[None, JWTPrincipal]:
        try:
            return self.__class__.jwt_principal(app=self.app).load(
                self.app.request.environ[self.header_token_key]
            )

        except jwt.ExpiredSignatureError:
            raise HTTPUnauthorized("Access token expired")

        except jwt.InvalidTokenError:
            pass

    def authenticate_request(self):
        self.app.identity = None
        if self.header_token_key not in self.app.request.environ:
            return

        encoded_token = self.app.request.environ[self.header_token_key]
        if encoded_token is None or not encoded_token.strip():
            return

        self.app.identity = self.get_principal()
        if not self.app.identity:
            raise HTTPBadRequest



from sqlalchemy import Column, between
from sqlalchemy.orm import Query

from cytra.db.access_control_mixin import AccessControlMixin
from cytra.db.inspection_mixin import InspectionMixin
from cytra.db.transform_mixin import TransformMixin
from cytra.exceptions import InvalidParamError


class FilteringMixin(TransformMixin, AccessControlMixin, InspectionMixin):
    """
    Filter a Sqlalchemy query by a `dict`

    Operators:

    - `=`      Exact string compare, escape the operators described below.
    - `!`      Not
    - `null`   Is null
    - `!null`  Is not null
    - `true`   Is True
    - `false`  Is False
    - `>=`     Greater than or Equal
    - `<=`     Less than or Equal
    - `>`      Greater than
    - `<`      Less than
    - `%~`     Case insensitive Like
    - `%`      Like
    - `^`      In, e.g: ^1,2,3
    - `!^`     Not In, e.g: !^1,2,3
    - `~`      Between,
               e.g: ~2018-05-29T11:23:16+04:30,2018-07-30T11:23:16+04:30

    Note: Without any operator means to Equal with keyword.
    Note: `LIKE` is case-insensitive by default in Sqlite

    Examples:

    where `name` is equal to `John`
    dict(name="John")

    where `age` is between `1` and `18`
    dict(age='~1,18')
    """  # noqa: E501

    @classmethod
    def _get_boolean_expr(cls, column):
        if hasattr(column, "is_"):
            return column

        if hasattr(column, "expr"):
            return column.expr(cls)

    @classmethod
    def filter_by_dict(cls, query: Query, criteria: dict) -> Query:
        for ic in cls.get_readables():
            if ic.type == "relationships":
                continue
            json_name = cls.export_column_name(
                col_key=ic.key, col_info=ic.info
            )
            if json_name in criteria:
                value = criteria[json_name]
                query = cls._filter_by_column_value(query, ic.column, value)

        return query

    @classmethod
    def filter_by_request(cls, query: Query) -> Query:
        return cls.filter_by_dict(query, cls.__app__.request.form)

    @classmethod
    def _filter_by_column_value(
        cls, query: Query, column: Column, value: str
    ) -> Query:
        import_value = getattr(cls, "import_value")
        if not isinstance(value, str):
            raise InvalidParamError

        if value.startswith("="):
            expr = column == import_value(column, value[1:])

        elif value.startswith("^") or value.startswith("!^"):
            value = value.split(",")
            not_ = value[0].startswith("!^")
            # flake8:noqa
            first_item = value[0][2 if not_ else 1 :]
            items = [first_item] + value[1:]
            items = [i for i in items if i.strip()]
            if not len(items):
                raise InvalidParamError("Invalid query string: %s" % value)
            expr = column.in_([import_value(column, j) for j in items])
            if not_:
                expr = ~expr

        elif value.startswith("~"):
            values = value[1:].split(",")
            start, end = [import_value(column, v) for v in values]
            expr = between(column, start, end)

        elif value == "null":
            expr = column.is_(None)
        elif value == "!null":
            expr = column.isnot(None)
        elif value == "!true":
            expr = column.isnot(True)
        elif value == "!false":
            expr = column.isnot(False)
        elif value.startswith("!"):
            expr = column != import_value(column, value[1:])
        elif value.startswith(">="):
            expr = column >= import_value(column, value[2:])
        elif value.startswith(">"):
            expr = column > import_value(column, value[1:])
        elif value.startswith("<="):
            expr = column <= import_value(column, value[2:])
        elif value.startswith("<"):
            expr = column < import_value(column, value[1:])
        elif value.startswith("%~"):
            expr = column.ilike("%%%s%%" % import_value(column, value[2:]))
        elif value.startswith("%"):
            expr = column.like("%%%s%%" % import_value(column, value[1:]))
        else:
            value = import_value(column, value)

            if value is True or value == "true":
                expr = cls._get_boolean_expr(column).is_(True)

            elif value is False or value == "false":
                expr = cls._get_boolean_expr(column).is_(False)

            else:
                expr = column == value

        return query.filter(expr)



from sqlalchemy import Column
from sqlalchemy.orm import composite as sa_composite
from sqlalchemy.orm import relationship as sa_relationship
from sqlalchemy.orm import synonym as sa_synonym


class Field(Column):
    """
    An overridden class from ``sqlalchemy.Column`` to apply ``cytra``
     properties.
    """

    inherit_cache = Column.inherit_cache

    def __init__(
        self,
        *args,
        dict_key: str = None,
        readonly: bool = None,
        protected: bool = None,
        info: dict = None,
        **kwargs
    ):
        """
        Initialize the field

        :param args: Positional-arguments that directly pass into
            ``sqlalchemy.Column.__init__``
        :param dict_key: Custom dictionary key related to this field,
            as default it will reads the field name and format
            (using ``cytra.BaseModel.__transformer__``) it before
            export.
        :param readonly: Make field read-only, it's mean this field will not
            accept any value from
            ``cytra.BaseModel.update_from_dict`` input dictionary.
        :param protected: Make field protected to representation
        :param info: Pass into Column info
        :param kwargs: Keyword-arguments that directly pass into
            ``sqlalchemy.Column.__init__``
        """
        info = info or dict()

        if dict_key is not None:
            info["dict_key"] = dict_key

        if readonly is not None:
            info["readonly"] = readonly

        if protected is not None:
            info["protected"] = protected

        super(Field, self).__init__(*args, info=info, **kwargs)


def relationship(
    *args, dict_key: str = None, protected: bool = None, **kwargs
):
    """
    Same as ``sqlalchemy.orm.relationship`` with extra arguments to use in
    ``cytra``.

    :param args: Positional-arguments that directly pass into
        ``sqlalchemy.orm.relationship``.
    :param dict_key: Custom dictionary key. default is formatted
        (using ``cytra.BaseModel.__transformer__``) attribute name
        (where ``relationship`` called).
    :param protected: Make field protected to representation.
    :param kwargs: Keyword-arguments that directly pass into
        ``sqlalchemy.orm.relationship``.
    :return:
    """
    info = dict()

    if dict_key is not None:
        info["dict_key"] = dict_key

    if protected is not None:
        info["protected"] = protected

    return sa_relationship(*args, info=info, **kwargs)


def composite(
    *args,
    dict_key: str = None,
    protected: bool = None,
    readonly: bool = None,
    **kwargs
):
    """
    Same as ``sqlalchemy.orm.composite`` with extra arguments to use in
    ``cytra``.

    :param args: Positional-arguments that directly pass into
        ``sqlalchemy.orm.composite``.
    :param dict_key: Custom dictionary key. default is formatted
        (using ``cytra.BaseModel.__transformer__``) attribute name
        (where ``composite`` called).
    :param protected:  Make field protected to representation.
    :param readonly: Make field read-only, it's mean this field will not accept
        any value from ``cytra.BaseModel.update_from_dict`` input
        dictionary.
    :param kwargs: Keyword-arguments that directly pass into
        ``sqlalchemy.orm.composite``.
    :return:
    """
    info = dict()

    if dict_key is not None:
        info["dict_key"] = dict_key

    if protected is not None:
        info["protected"] = protected

    if readonly is not None:
        info["readonly"] = readonly

    return sa_composite(*args, info=info, **kwargs)


def synonym(
    *args,
    dict_key: str = None,
    protected: bool = None,
    readonly: bool = None,
    **kwargs
):
    """
    Same as ``sqlalchemy.orm.synonym`` with extra arguments to use in
    ``cytra``.

    .. versionadded:: 0.4.0

    :param args: Positional-arguments that directly pass into
        ``sqlalchemy.orm.synonym``.
    :param dict_key: Custom dictionary key. default is formatted
        (using ``cytra.BaseModel.__transformer__``) attribute name
        (where ``synonym`` called).
    :param protected:  Make field protected to representation.
    :param readonly: Make field read-only, it's mean this field will not accept
        any value from ``cytra.BaseModel.update_from_dict`` input
        dictionary.
    :param kwargs: Keyword-arguments that directly pass into
        ``sqlalchemy.orm.synonym``.
    :return:
    """
    info = dict()

    if dict_key is not None:
        info["dict_key"] = dict_key

    if protected is not None:
        info["protected"] = protected

    if readonly is not None:
        info["readonly"] = readonly

    return sa_synonym(*args, info=info, **kwargs)



from datetime import date, datetime, time
from decimal import Decimal
from typing import Any, Union, get_type_hints

import pytz
from sqlalchemy import Column
from sqlalchemy.ext.hybrid import HYBRID_METHOD, HYBRID_PROPERTY
from sqlalchemy.orm import CompositeProperty
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.relationships import RelationshipProperty

from cytra.constants import (
    ISO_DATE_FORMAT,
    ISO_DATETIME_FORMAT,
    ISO_DATETIME_PATTERN,
    ISO_TIME_FORMAT,
)
from cytra.helpers import to_camel_case


class Transformer:
    """Model transformer abstract class"""

    @classmethod
    def export_column_name(cls, name):
        """
        Export column name
        :param name:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def export_datetime(cls, value: datetime):
        """
        Export python datetime
        :param value:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def export_date(cls, value: date):
        """
        Export python date
        :param value:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def export_time(cls, value: time):
        """
        Export python time
        :param value:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def import_datetime(cls, value: Union[str, int]) -> datetime:
        """
        Import datetime field
        :param value:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def import_date(cls, value: Union[str, int]) -> date:
        """
        Import date
        :param value:
        :return:
        """
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def import_time(cls, value: Union[str, int]) -> time:
        """
        Import time
        :param value:
        :return:
        """
        raise NotImplementedError  # pragma: no cover


class DefaultTransformer(Transformer):
    @classmethod
    def export_column_name(cls, key):
        return to_camel_case(key)

    @classmethod
    def export_datetime(cls, value):
        return value.isoformat()

    @classmethod
    def export_date(cls, value):
        return value.isoformat()

    @classmethod
    def export_time(cls, value):
        return value.isoformat()

    @classmethod
    def import_datetime(cls, value):
        match = ISO_DATETIME_PATTERN.match(value)
        if not match:
            raise ValueError("Invalid datetime format")

        res = datetime.strptime(match.group(1), ISO_DATETIME_FORMAT)
        if match.group(2) and len(match.group(2)) > 0:
            res = res.replace(microsecond=int(match.group(2)))

        return res

    @classmethod
    def import_date(cls, value):
        try:
            return datetime.strptime(value, ISO_DATE_FORMAT).date()
        except ValueError:
            raise ValueError("Invalid date format")

    @classmethod
    def import_time(cls, value):
        try:
            return datetime.strptime(value, ISO_TIME_FORMAT).time()
        except ValueError:
            raise ValueError("Invalid date format")


class DefaultTransformerUTC(DefaultTransformer):
    @classmethod
    def export_datetime(cls, value):
        value = value.replace(tzinfo=pytz.utc)
        return super().export_datetime(value)

    @classmethod
    def export_time(cls, value):
        value = value.replace(tzinfo=pytz.utc)
        return super().export_time(value)


class TransformMixin:
    """
    Transform columns

    Note: All mixins must respect to this mixin while intracting with
    outside of the model.
    """

    __transformer__ = DefaultTransformerUTC

    @classmethod
    def import_value(cls, c: Column, v: Union[None, str]) -> Any:
        """
        Import value for a column.
        :param column:
        :param v:
        :return:
        """
        if v is None:
            return v

        if isinstance(c, Column) or isinstance(c, InstrumentedAttribute):
            try:
                type_ = c.type.python_type
            except NotImplementedError:
                type_ = None

        elif hasattr(c, "descriptor"):
            # using type annotation for hybrid_property
            cd = c.descriptor
            if (
                hasattr(cd, "extension_type")
                and cd.extension_type == HYBRID_PROPERTY
            ):
                type_ = get_type_hints(cd.fget).get("return")
            else:
                type_ = None

        elif (
            hasattr(c, "extension_type")
            and c.extension_type == HYBRID_PROPERTY
        ):
            type_ = get_type_hints(c.fget).get("return")

        else:  # pragma: nocover
            type_ = None

        if type_ is bool and not isinstance(v, bool):
            return str(v).lower() == "true"

        if type_ == datetime:
            return cls.__transformer__.import_datetime(v)

        if type_ == date:
            return cls.__transformer__.import_date(v)

        if type_ == time:
            return cls.__transformer__.import_time(v)

        return v

    @classmethod
    def export_value(cls, c: Column, v: Any) -> tuple:
        """
        Prepare column value to export.
        :param c:
        :param v:
        :return:
        """
        if (
            hasattr(c, "property")
            and isinstance(c.property, RelationshipProperty)
            and c.property.uselist
        ):
            return [c.to_dict() for c in v]

        if hasattr(c, "property") and isinstance(
            c.property, CompositeProperty
        ):
            return v.__composite_values__()

        if v is None:
            return v

        if isinstance(v, datetime):
            return cls.__transformer__.export_datetime(v)

        if isinstance(v, date):
            return cls.__transformer__.export_date(v)

        if isinstance(v, time):
            return cls.__transformer__.export_time(v)

        if hasattr(v, "to_dict"):
            return v.to_dict()

        if isinstance(v, Decimal):
            return str(v)

        if c.extension_type == HYBRID_METHOD:
            return v(cls)

        return v

    @classmethod
    def export_column_name(cls, col_key: str, col_info: dict) -> str:
        return cls.__transformer__.export_column_name(
            col_info.get("dict_key", col_key)
        )



from typing import Union

from gongish import HTTPNoContent


class CORS:
    """
    Cross-Origin Resource Sharing (CORS) is an HTTP-header based mechanism
    that allows a server to indicate any origins (domain, scheme, or port)
    other than its own from which a browser should permit loading of resources.
    CORS also relies on a mechanism by which browsers make a "preflight"
    request to the server hosting the cross-origin resource, in order to check
    that the server will permit the actual request. In that preflight,
    the browser sends headers that indicate the HTTP method and headers that
    will be used in the actual request.
    https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS

    """

    allow_origin: Union[None, str] = None
    """
    The Access-Control-Allow-Origin response header indicates whether
    the response can be shared with requesting code from the given origin.

    Directives:
    *: For requests without credentials, the literal value "*" can
        be specified, as a wildcard; the value tells browsers to allow
        requesting code from any origin to access the resource.
        Attempting to use the wildcard with credentials will result
        in an error.
    <origin>: Specifies an origin. Only a single origin can be specified.
        If the server supports clients from multiple origins, it must return
        the origin for the specific client making the request.
    """

    allow_methods: Union[None, set, str] = None
    """
    The Access-Control-Allow-Methods response header specifies the method or
    methods allowed when accessing the resource in response to a preflight
    request.

    Directives:
    set: list of the allowed HTTP request methods.
    *: The value "*" only counts as a special wildcard value for requests
        without credentials (requests without HTTP cookies or
        HTTP authentication information).
        In requests with credentials, it is treated as the literal method name
        "*" without special semantics.
    """

    allow_headers: Union[None, set, str] = None
    """
    The Access-Control-Allow-Headers response header is used in response to
    a preflight request which includes the Access-Control-Request-Headers to
    indicate which HTTP headers can be used during the actual request.

    Directives:
    set: list of the allowed HTTP request headers.
    *: The value "*" only counts as a special wildcard value for requests
        without credentials (requests without HTTP cookies or
        HTTP authentication information). In requests with credentials,
        it is treated as the literal header name "*" without special semantics.
        Note that the Authorization header can't be wildcarded and always needs
        to be listed explicitly.
    """

    expose_headers: Union[None, set, str] = None
    """
    The Access-Control-Expose-Headers response header allows a server to
    indicate which response headers should be made available to scripts running
    in the browser, in response to a cross-origin request.

    Directives:
    set: A list of header names that clients are allowed to access from a
        response.
        These are in addition to the CORS-safelisted response headers.

    *: The value "*" only counts as a special wildcard value for
        requests without credentials (requests without HTTP cookies or
        HTTP authentication information).
        In requests with credentials, it is treated as the literal header
        name "*" without special semantics. Note that the Authorization header
        can't be wildcarded and always needs to be listed explicitly.
    """

    allow_credentials: Union[None, bool] = None
    """
    Indicates whether or not the response to the request can be exposed when
    the credentials flag is true.
    """

    max_age: Union[None, int] = None
    """
    The Access-Control-Max-Age response header indicates how long the results
    of a preflight request (that is the information contained in the
    Access-Control-Allow-Methods and Access-Control-Allow-Headers headers) can
    be cached.

    Directives:
    Maximum number of seconds the results can be cached.
    Firefox caps this at 24 hours (86400 seconds).
    Chromium (prior to v76) caps at 10 minutes (600 seconds).
    Chromium (starting in v76) caps at 2 hours (7200 seconds).
    Chromium also specifies a default value of 5 seconds.
    A value of -1 will disable caching, requiring a preflight OPTIONS check
    for all calls.
    """

    def __init__(self) -> None:
        self.allow_origin = None
        self.allow_methods = set()
        self.allow_headers = set()
        self.allow_credentials = None
        self.expose_headers = set()
        self.max_age = None

    @property
    def headers(self):
        res = dict()
        if self.allow_origin is not None:
            res["access-control-allow-origin"] = self.allow_origin

        if self.allow_methods is not None:
            res["access-control-allow-methods"] = (
                ",".join(self.allow_methods).upper()
                if isinstance(self.allow_methods, set)
                else self.allow_methods
            )

        if self.allow_headers is not None:
            res["access-control-allow-headers"] = (
                ",".join(self.allow_headers)
                if isinstance(self.allow_headers, set)
                else self.allow_headers
            )

        if self.allow_credentials is True:
            res["access-control-allow-credentials"] = "true"

        if self.expose_headers is not None:
            res["access-control-expose-headers"] = (
                ",".join(self.expose_headers)
                if isinstance(self.expose_headers, set)
                else self.expose_headers
            )

        return res


class CORSAppMixin:
    def _set_cors(self):
        self.response.headers.update(self._cors_headers)

    def dispatch(self, path, verb):
        if verb == "options":

            def handler():
                raise HTTPNoContent

            return handler, ()

        return super().dispatch(path, verb)

    def on_begin_response(self):
        self._set_cors()
        return super().on_begin_response()

    def on_error(self, exc):
        self._set_cors()
        return super().on_error(exc)

    def setup(self):
        super().setup()
        self.cors.allow_methods.update(self.verbs)
        self._cors_headers = self.cors.headers
        self.verbs.add("options")

    def shutdown(self):
        super().shutdown()
        self._cors_headers = None
        self.verbs.remove("options")



from typing import Any, NamedTuple

from sqlalchemy import Column, inspect
from sqlalchemy.ext.associationproxy import ASSOCIATION_PROXY
from sqlalchemy.ext.hybrid import HYBRID_PROPERTY


class InspectedColumn(NamedTuple):
    type: str
    key: str
    info: dict
    column: Any


class InspectionMixin:
    __abstract__ = True
    __cytra_inspections__: dict = None
    __cytra_inspections_key__: str = None

    @classmethod
    def _update_inspections(cls):
        res = dict(
            all=[],
            all_by_key={},
            hybrid_properties=[],
            relationships=[],
            composites=[],
            synonyms=[],
            columns=[],
            proxies=[],
        )
        mapper = inspect(cls)
        for k, c in mapper.all_orm_descriptors.items():

            if k == "__mapper__":  # pragma:nocover
                continue

            if c.extension_type == ASSOCIATION_PROXY:
                cc_type = "proxies"

            elif c.extension_type == HYBRID_PROPERTY:
                cc_type = "hybrid_properties"

            elif k in mapper.relationships:
                cc_type = "relationships"

            elif k in mapper.synonyms:
                cc_type = "synonyms"

            elif k in mapper.composites:
                cc_type = "composites"

            else:
                cc_type = "columns"

            ic = InspectedColumn(
                type=cc_type,
                key=cls.get_column_key(c),
                info=cls.get_column_info(c),
                column=c,
            )
            res[cc_type].append(ic)
            res["all"].append(ic)
            res["all_by_key"][ic.key] = ic
        return res

    @classmethod
    def get_inspections(cls):
        key = cls.__name__
        if cls.__cytra_inspections__ is None or (
            cls.__cytra_inspections_key__ != key
        ):
            cls.__cytra_inspections__ = cls._update_inspections()
            cls.__cytra_inspections_key__ = key

        return cls.__cytra_inspections__

    @classmethod
    def get_column_info(cls, column: Column) -> dict:
        # Use original property for proxies
        if hasattr(column, "original_property") and column.original_property:
            info = column.info.copy()
            info.update(column.original_property.info)
            return info

        return column.info

    @classmethod
    def get_column_key(cls, column: Column) -> str:
        if hasattr(column, "key"):
            return column.key

        # `hybrid_method`
        if hasattr(column, "func"):
            return column.func.__name__

        # `hybrid_property`
        return column.__name__



from sqlalchemy.orm import Query

from cytra.exceptions import PaginationLimitError


class PaginationMixin:
    __take_header_key__ = "HTTP_X_TAKE"
    __skip_header_key__ = "HTTP_X_SKIP"
    __max_take__ = 100

    @classmethod
    def paginate_by_request(cls, query: Query):
        app = cls.__app__
        qs = app.request.query
        env = app.request.environ

        try:
            take = int(qs.get("take") or env.get(cls.__take_header_key__))
            if take > cls.__max_take__:
                raise PaginationLimitError
        except (ValueError, TypeError):
            take = cls.__max_take__

        try:
            skip = int(qs.get("skip") or env.get(cls.__skip_header_key__))
        except (ValueError, TypeError):
            skip = 0

        headers = app.response.headers
        headers["x-pagination-take"] = str(take)
        headers["x-pagination-skip"] = str(skip)
        headers["x-pagination-count"] = str(query.count())
        return query.offset(skip).limit(take)



from sqlalchemy import desc
from sqlalchemy.orm import Query

from cytra.db.access_control_mixin import AccessControlMixin
from cytra.db.transform_mixin import TransformMixin


class OrderingMixin(TransformMixin, AccessControlMixin):
    """
    Apply ordering to query from string expression

    purposed to use in URL/query-string.

    for example:
    col1,col2,-col3

    equivalent to:
    asc(col1) then asc(col2) then desc(col3)
    """

    @classmethod
    def create_sort_criteria(cls, sort_columns):
        valid_columns = {
            cls.export_column_name(ic.key, ic.info): getattr(cls, ic.key)
            for ic in cls.get_readables()
            if ic.type != "relationships"
        }
        for column_name, column_is_descending in sort_columns:
            if column_name in valid_columns:
                yield (valid_columns[column_name], column_is_descending)

    @classmethod
    def _sort_by_key_value(cls, query, column, descending=False):
        expression = column

        if column.info.get("collation"):
            expression = expression.collate(column.info["collation"])

        if descending:
            expression = desc(expression)

        return query.order_by(expression)

    @classmethod
    def sort_query(cls, query: Query, sort_exp: dict) -> Query:
        sort_columns = [
            (c[1:] if c.startswith("-") else c, c.startswith("-"))
            for c in sort_exp.split(",")
        ]

        criteria = tuple(cls.create_sort_criteria(sort_columns))

        for criterion in criteria:
            query = cls._sort_by_key_value(query, *criterion)

        return query

    @classmethod
    def sort_by_request(cls, query: Query) -> Query:
        return cls.sort_query(query, cls.__app__.request.query.get("sort", ""))



from setuptools import find_packages, setup


def read_version(module_name):
    from os.path import dirname, join
    from re import S, match

    with open(join(dirname(__file__), module_name, "__init__.py")) as f:
        return match(r".*__version__.*('|\")(.*?)('|\")", f.read(), S).group(2)


setup(
    name="cytra",
    version=read_version("cytra"),
    keywords="database",
    packages=find_packages(),
    install_requires=[
        "gongish >= 1.1.0",
        "sqlalchemy >= 1.4.0",
        "pytz >= 2021.1",
        "PyJWT >= 2.1.0",
        "redis >= 3.5.3",
        "user-agents >= 2.2.0",
        "webtest >= 2.0.35",
    ],
    extras_require=dict(
        ujson=["ujson >= 4.0.0"],
    ),
)



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Mapper, Query, Session
from sqlalchemy.sql.schema import MetaData

from cytra.db import (
    FilteringMixin,
    OrderingMixin,
    PaginationMixin,
    SerializeMixin,
)


class BaseModel(SerializeMixin):
    __app__ = None

    @classmethod
    def compose_query(cls, query: Query) -> Query:
        if issubclass(cls, FilteringMixin):
            query = cls.filter_by_request(query)

        if issubclass(cls, OrderingMixin):
            query = cls.sort_by_request(query)

        if issubclass(cls, PaginationMixin):
            query = cls.paginate_by_request(query)

        return query


class CytraDBQuery(Query):
    __app__ = None
    _cytra_target: Mapper = None

    def expose(self) -> list:
        return self._cytra_target.dump_query(
            self._cytra_target.compose_query(query=self)
        )

    def __init__(self, entities, session=None):
        firstentity = entities[0]
        if isinstance(firstentity, Mapper):
            self._cytra_target = firstentity.entity

        super().__init__(entities, session)


class DBSessionProxy(object):
    __cytra_session__ = None

    def __new__(cls) -> Session:
        return super().__new__(cls)

    def __getattr__(self, key):
        return getattr(self.__class__.__cytra_session__, key)

    def __setattr__(self, key, value):
        setattr(self.__class__.__cytra_session__, key, value)

    def __delattr__(self, key):
        delattr(self.__class__.__cytra_session__, key)


metadata = MetaData()
DeclarativeBase: BaseModel = declarative_base(cls=BaseModel, metadata=metadata)
dbsession = DBSessionProxy()



from gongish.exceptions import HTTPInternalServerError
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session, scoped_session, sessionmaker

from cytra.db.base import CytraDBQuery, DBSessionProxy, DeclarativeBase
from cytra.exceptions import PostgresSQLError


class DatabaseAppMixin:
    db: Session = None

    def on_end_response(self):
        super().on_end_response()
        if self.db:
            self.db.remove()

    def handle_exception(self, exc, start_response):
        if isinstance(exc, SQLAlchemyError):
            self._log.exception(str(exc))
            if PostgresSQLError.validate_postgres_error(exc):
                exc = PostgresSQLError(exc)
            else:
                exc = HTTPInternalServerError()
        return super().handle_exception(exc, start_response)

    def setup(self):
        super().setup()
        if "sqlalchemy" in self.config:
            db_config = self.config.sqlalchemy
            session_kw = dict(db_config.get("session", dict()))
            session_factory = sessionmaker(
                bind=create_engine(**db_config.engine),
                query_cls=CytraDBQuery,
                **session_kw
            )
            session = scoped_session(session_factory)
            DeclarativeBase.query = session.query_property()
            DeclarativeBase.__app__ = self
            CytraDBQuery.__app__ = self
            self.db = session
            DBSessionProxy.__cytra_session__ = session
            self.cors.expose_headers.update(
                (
                    "x-pagination-count",
                    "x-pagination-skip",
                    "x-pagination-take",
                )
            )

        else:
            self.db = None

    def shutdown(self):
        if self.db is not None:
            self.db.close()
            self.db.get_bind().dispose()
            DeclarativeBase.__app__ = None
            CytraDBQuery.__app__ = None
            DBSessionProxy.__cytra_session__ = None

        super().shutdown()

    def commit(self):
        try:
            self.db.commit()
        except SQLAlchemyError:
            self.db.rollback()
            raise



import jwt
from gongish import HTTPForbidden


class JWTPrincipal:
    """
    Role based JWT Principal
    """

    def __init__(self, app, payload=None):
        self.app = app
        self.payload = payload

    def dump(self) -> str:
        config = self.app.config
        return jwt.encode(
            payload=self.payload,
            key=config.auth.jwt_secret_key,
            algorithm=config.auth.jwt_algorithm,
        )

    def load(self, token: str):
        config = self.app.config
        self.payload = jwt.decode(
            token,
            key=config.auth.jwt_secret_key,
            algorithms=config.auth.jwt_algorithm,
        )
        return self

    def to_dict(self):
        return self.payload

    @property
    def roles(self):
        return self.payload.get("roles")

    def is_in_roles(self, *roles):
        if self.payload.get("roles"):
            if set(self.payload["roles"]).intersection(roles):
                return True
        return False

    def assert_roles(self, *roles):
        if roles and not self.is_in_roles(*roles):
            raise HTTPForbidden

    @property
    def id(self):
        return int(self.payload.get("id"))

    @property
    def session_id(self):
        return self.payload.get("sessionId")


class DefaultJWTPrincipal(JWTPrincipal):
    @property
    def email(self):
        return self.payload.get("email")

    @property
    def mobile(self):
        return self.payload.get("mobile")



from typing import List

from sqlalchemy.orm import Query

from cytra.db.access_control_mixin import AccessControlMixin
from cytra.db.transform_mixin import TransformMixin


class SerializeMixin(TransformMixin, AccessControlMixin):
    def to_dict(self) -> dict:
        result = dict()
        for ic in self.__class__.get_readables(self):
            key = self.get_column_key(ic.column)
            result.setdefault(
                self.export_column_name(key, ic.info),
                self.export_value(ic.column, getattr(self, key)),
            )
        return result

    def update_from_dict(self, data: dict) -> None:
        datakeys = data.keys()
        for ic in self.__class__.get_writables(self):
            key = self.get_column_key(ic.column)
            key_ = self.export_column_name(key, ic.info)
            if key_ not in datakeys:
                continue

            setattr(self, key, self.import_value(ic.column, data[key_]))

    def update_from_request(self) -> None:
        self.update_from_dict(self.__app__.request.form)

    @classmethod
    def dump_query(cls, query: Query) -> List[dict]:
        """
        Dump query results in a list of model dictionaries.
        :param query:
        :return:
        """
        return [o.to_dict() for o in query]



import os
from os.path import exists
from urllib.parse import urlparse

from sqlalchemy import create_engine


class DatabaseManagerAbstract(object):
    def __init__(self, engine_kwargs, admin_engine_kwargs=None):
        self.db_url = engine_kwargs["url"]
        self.db_name = urlparse(self.db_url).path.lstrip("/")
        if admin_engine_kwargs:
            self.engine = create_engine(**admin_engine_kwargs)
        else:
            self.engine = create_engine(**engine_kwargs)

    def __enter__(self):
        self.connection = self.engine.connect()
        self.connection.execute("commit")
        return self

    def __exit__(self, *args):
        self.engine.dispose()

    def create_database_if_not_exists(self):
        if not self.database_exists():
            self.create_database()

    def database_exists(self):  # pragma: no cover
        raise NotImplementedError()

    def create_database(self):  # pragma: no cover
        raise NotImplementedError()

    def drop_database(self):  # pragma: no cover
        raise NotImplementedError()


class SqliteManager(DatabaseManagerAbstract):
    """
    SQLite database manager
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = self.db_url.replace("sqlite:///", "")

    def database_exists(self):
        return exists(self.filename)

    def create_database(self):
        if self.database_exists():
            raise RuntimeError(
                "The file is already exists: %s" % self.filename
            )
        open(self.filename, "a").close()

    def drop_database(self):
        os.remove(self.filename)


class MysqlManager(DatabaseManagerAbstract):
    """
    MySQL database manager
    """

    def database_exists(self):
        r = self.connection.execute(f"SHOW DATABASES LIKE '{self.db_name}'")
        try:
            ret = r.cursor.fetchall()
            return len(ret) > 0
        finally:
            r.cursor.close()

    def create_database(self):
        self.connection.execute(f"CREATE DATABASE {self.db_name}")
        self.connection.execute("commit")

    def drop_database(self):
        self.connection.execute(f"DROP DATABASE IF EXISTS {self.db_name}")
        self.connection.execute("commit")


class PostgresManager(DatabaseManagerAbstract):
    def database_exists(self):
        r = self.connection.execute(
            f"SELECT 1 FROM pg_database WHERE datname = '{self.db_name}'"
        )
        try:
            ret = r.cursor.fetchall()
            return ret
        finally:
            r.cursor.close()

    def create_database(self):
        self.connection.execute(f"CREATE DATABASE {self.db_name}")
        self.connection.execute("commit")

    def drop_database(self):
        self.connection.execute(f"DROP DATABASE IF EXISTS {self.db_name}")
        self.connection.execute("commit")


class DatabaseManager(DatabaseManagerAbstract):
    """
    Some operations to manage database
    """

    def __new__(cls, engine_kwargs, admin_engine_kwargs=None):
        url = engine_kwargs["url"]
        if url.startswith("sqlite"):
            manager_class = SqliteManager
        elif url.startswith("postgres"):
            manager_class = PostgresManager
        elif url.startswith("mysql"):
            manager_class = MysqlManager
        else:
            raise ValueError("Unsupported database uri: %s" % url)

        return manager_class(engine_kwargs, admin_engine_kwargs)



import gettext
import locale as lib_locale
import logging

logger = logging.getLogger("cytra")


def get_request_locales(headers):
    # e.g. 'es,en-US;q=0.8,en;q=0.6'
    locale_header = headers.get("accept-language", "").strip()
    if not locale_header:
        return

    if ";q=" in locale_header:
        # Extract all locales and their preference (q)
        locales = set()  # e.g. [('es', 1.0), ('en-US', 0.8), ('en', 0.6)]
        for locale_str in locale_header.split(","):
            locale_parts = locale_str.split(";q=")
            locales.add(
                (
                    locale_parts[0],
                    float(locale_parts[1]) if len(locale_parts) > 1 else 1.0,
                )
            )
        locales = map(
            lambda x: x[0],
            sorted(locales, key=lambda x: x[1], reverse=True),
        )
    else:
        locales = locale_header.split(",")

    # Sort locales according to preference
    for locale in locales:
        parts = locale.strip().split("-")
        if len(parts) == 1:
            yield parts[0], None
        else:
            yield parts[0], parts[1].upper()


class LocaleManager:
    def __init__(self, app, locales, domain, localedir):
        self.app = app
        self.locales = tuple(locales)
        self.locales_tuple = tuple(map(lambda x: tuple(x.split("_")), locales))
        self.translations = {}
        self.default = locales[0]
        self.set_locale(self.default)

        # load translations
        for locale in locales:
            self.translations[locale] = gettext.translation(
                domain=domain,
                localedir=localedir,
                languages=[locale],
            )

    def find_locale_from_request(self):
        request_locales = tuple(get_request_locales(self.app.request.headers))
        if request_locales:
            # find exact match
            for locale in request_locales:
                for locale_ in self.locales_tuple:
                    if locale == locale_:
                        return locale_

            # find by language match
            for lang, region in request_locales:
                for lang_, region_ in self.locales_tuple:
                    if lang == lang_:
                        return lang_, region_

        return self.default.split("_")

    def set_locale_from_request(self):
        lang, region = self.find_locale_from_request()
        self.set_locale(f"{lang}_{region}")
        self.app.response.headers["content-language"] = f"{lang}-{region}"

    def set_locale(self, locale):
        self.locale = locale
        try:
            lib_locale.setlocale(lib_locale.LC_ALL, locale)
        except lib_locale.Error:
            logger.exception("Locale error (%s)" % locale)

    def translate(self, word, plural=None, n=None):
        translation = self.translations[self.locale]
        if plural is not None:
            return translation.ngettext(word, plural, n)
        return translation.gettext(word)


class I18nAppMixin:
    """
    Internationalization Mixin

    Configuration:

    Example configuration in YAML:
        i18n:
          locales:
            - en_US
            - fa_IR
          localedir: myapp/i18n
          domain: app

    Note: First locale will set as default.

    Change locale:
        >>> app.i18n.set_locale('en_US')

    Translate:
        >>> app.i18n.translate('HelloWorld')
    """

    i18n = None

    def setup(self):
        super().setup()
        if "i18n" not in self.config:  # pragma: nocover
            return

        self.i18n = LocaleManager(self, **dict(self.config.i18n))
